/*-
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* for strdup() */
//#define _POSIX_C_SOURCE 200809L

#include "config.h"

#include <stdlib.h>
#include <string.h>
#include <brotli/encode.h>

#include "cache/cache.h"
#include "cache/cache_filter.h"

#include "vcc_if.h"

/* XXX make this configurable; cf. varnishd default gzip_buffer */
#define BUFFER_SZ (32 * 1024 * 1024)

struct vbr_enc {
	unsigned		magic;
#define VBR_ENC_MAGIC		0x467c387b
	BrotliEncoderState	*state;
	uint8_t			*buf;
	const uint8_t		*next_in;
	uint8_t			*next_out;
	size_t			bufsz;
	size_t			buflen;
	size_t			avail_in;
	size_t			avail_out;
	size_t			total_out;
};

static struct vbr_enc *
newEnc(void)
{
	struct vbr_enc *vbr_enc;

	/* XXX inc encs counter */
	ALLOC_OBJ(vbr_enc, VBR_ENC_MAGIC);
	if (vbr_enc == NULL)
		return (NULL);
	vbr_enc->state = BrotliEncoderCreateInstance(NULL, NULL, NULL);
	if (vbr_enc->state == NULL)
		return (NULL);
	/* XXX configure the handle */
	return (vbr_enc);
}

static void
destroy(struct vbr_enc **vp)
{
	struct vbr_enc *vbr_enc;

	TAKE_OBJ_NOTNULL(vbr_enc, vp, VBR_ENC_MAGIC);
	/* XXX log results: bytes in, bytes out */
	/* XXX IMO no finish op necessary here */
	BrotliEncoderDestroyInstance(vbr_enc->state);
	if (vbr_enc->buf != NULL)
		free(vbr_enc->buf);
	FREE_OBJ(vbr_enc);
}

static int
getBuf(struct vbr_enc *vbr_enc)
{
	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);
	AZ(vbr_enc->bufsz);
	AZ(vbr_enc->buflen);
	AZ(vbr_enc->buf);

	/* XXX make this configurable */
	vbr_enc->bufsz = BUFFER_SZ;
	vbr_enc->buf = malloc(vbr_enc->bufsz);
	if (vbr_enc->buf == NULL) {
		vbr_enc->bufsz = 0;
		return (-1);
	}
	return (0);
}

static void
setInputBuf(struct vbr_enc *vbr_enc, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);

	AZ(vbr_enc->avail_in);
	vbr_enc->next_in = TRUST_ME(ptr);
	vbr_enc->avail_in = len;
}

static int
isInputBufEmpty(const struct vbr_enc *vbr_enc)
{
	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);
	return (vbr_enc->avail_in == 0);
}

static void
setOutputBuf(struct vbr_enc *vbr_enc, const void *ptr, ssize_t len)
{
	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);

	vbr_enc->next_out = TRUST_ME(ptr);
	vbr_enc->avail_out = len;
}

#if 0
static int
isOutputBufFull(const struct vbr_enc *vbr_enc)
{
	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);
	return (vbr_enc->avail_out == 0);
}
#endif

static BROTLI_BOOL
encode(struct vbr_enc *vbr_enc, ssize_t *dl, int finished)
{
	BROTLI_BOOL ret;
	const uint8_t *before;
	enum BrotliEncoderOperation op = BROTLI_OPERATION_PROCESS;

	CHECK_OBJ_NOTNULL(vbr_enc, VBR_ENC_MAGIC);
	before = vbr_enc->next_out;

	if (finished)
		op = BROTLI_OPERATION_FINISH;
	ret = BrotliEncoderCompressStream(vbr_enc->state, op,
					  &vbr_enc->avail_in,
					  &vbr_enc->next_in,
					  &vbr_enc->avail_out,
					  &vbr_enc->next_out,
					  &vbr_enc->total_out);
	*dl = (const uint8_t *)vbr_enc->next_out - before;
	return (ret);
}

/* VFP interface */

static enum vfp_status v_matchproto_(vfp_init_f)
vfp_brenc_init(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr_enc *vbr_enc;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	/* XXX: cache_gzip.c has:
	 if (http_GetStatus(vc->resp) == 206)
	 	return (VFP_NULL);
	 What do we want to do about partial responses?
	*/

	if (http_GetHdr(ctx->resp, H_Content_Encoding, NULL))
		return (VFP_NULL);
	if ((vbr_enc = newEnc()) == NULL)
		return (VFP_ERROR);
	ent->priv1 = vbr_enc;
	if (getBuf(vbr_enc))
		return (VFP_ERROR);
	setInputBuf(vbr_enc, vbr_enc->buf, 0);
	AZ(vbr_enc->buflen);

	http_Unset(ctx->resp, H_Content_Length);
	RFC2616_Weaken_Etag(ctx->resp);
	http_SetHeader(ctx->resp, "Content-Encoding: br");
	RFC2616_Vary_AE(ctx->resp);
	return (VFP_OK);
}

static enum vfp_status v_matchproto_(vfp_pull_f)
vfp_brenc_pull(struct vfp_ctx *ctx, struct vfp_entry *ent, void *ptr,
	       ssize_t *lenp)
{
	struct vbr_enc *vbr_enc;
	ssize_t len, dl;
	enum vfp_status vp = VFP_ERROR;
	int finished = 0;
	BROTLI_BOOL done;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);
	CAST_OBJ_NOTNULL(vbr_enc, ent->priv1, VBR_ENC_MAGIC);
	AN(ptr);
	AN(lenp);

	len = *lenp;
	*lenp = 0;
	setOutputBuf(vbr_enc, ptr, len);

	do {
		if (isInputBufEmpty(vbr_enc)) {
			len = vbr_enc->bufsz;
			vp = VFP_Suck(ctx, vbr_enc->buf, &len);
			if (vp == VFP_ERROR)
				break;
			if (vp == VFP_END)
				finished = 1;
			setInputBuf(vbr_enc, vbr_enc->buf, len);
		}
		if (!isInputBufEmpty(vbr_enc) || finished) {
			done = encode(vbr_enc, &dl, finished);
			if (done != BROTLI_TRUE)
				return (VFP_Error(ctx, "brotli encode failed"));
			if (dl > 0) {
				*lenp = dl;
				return (VFP_OK);
			}
		}
		AN(isInputBufEmpty(vbr_enc));
	} while (!finished);

	if (done != BROTLI_TRUE)
		return (VFP_Error(ctx, "brotli encode failed"));
	return (VFP_END);
}

static void v_matchproto_(vfp_fini_f)
vfp_brenc_fini(struct vfp_ctx *ctx, struct vfp_entry *ent)
{
	struct vbr_enc *vbr_enc;

	CHECK_OBJ_NOTNULL(ctx, VFP_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(ent, VFP_ENTRY_MAGIC);

	if (ent->priv1 != NULL) {
		CAST_OBJ(vbr_enc, ent->priv1, VBR_ENC_MAGIC);
		ent->priv1 = NULL;
		destroy(&vbr_enc);
	}
}

static const struct vfp vfp_brenc = {
	.name = "br",
	.init = vfp_brenc_init,
	.pull = vfp_brenc_pull,
	.fini = vfp_brenc_fini,
};

/* Event function */

int
vmod_event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e e)
{
	ASSERT_CLI();
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(priv);

	switch(e) {
	case VCL_EVENT_LOAD:
		VRT_AddVFP(ctx, &vfp_brenc);
		return (0);
	case VCL_EVENT_DISCARD:
		VRT_RemoveVFP(ctx, &vfp_brenc);
		return (0);
	case VCL_EVENT_WARM:
	case VCL_EVENT_COLD:
		return (0);
	default:
		WRONG("illegal event enum");
	}
	NEEDLESS(return (0));
}

static VCL_STRING
brotli_version(VRT_CTX, uint32_t version)
{
	VCL_STRING vers;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	unsigned major = version >> 24;
	unsigned minor = (version >> 12) & 0xFFF;
	unsigned patch = version & 0xFFF;
	if ((vers = WS_Printf(ctx->ws, "%u.%u.%u", major, minor, patch))
	    == NULL) {
		VSLb(ctx->vsl, SLT_VCL_Error, "vfp brotli error: "
		     "insufficient workspace for version string");
		return ("(insufficient workspace for version string)");
	}
	return vers;
}

VCL_STRING
vmod_encoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliEncoderVersion()));
}

#if 0
VCL_STRING
vmod_decoder_version(VRT_CTX)
{
	return (brotli_version(ctx, BrotliDecoderVersion()));
}
#endif

VCL_STRING
vmod_version(VRT_CTX)
{
	(void) ctx;
	return VERSION;
}
